/* "NETGEN", a netlist-specification tool for VLSI
   Copyright (C) 1989, 1990   Massimo A. Sivilotti
   Author's address: mass@csvax.cs.caltech.edu;
                     Caltech 256-80, Pasadena CA 91125.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (any version).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file copying.  If not, write to
the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* query.c -- simple command-line interpreter */

#include "config.h"

#include <stdio.h>
#include <ctype.h>
#include <setjmp.h>
#include <signal.h>

#ifdef HAVE_MALLINFO
#include <malloc.h>
#endif

#ifdef IBMPC
#include <alloc.h>
#include <process.h> /* for system() */
#include <bios.h>  /* for bioskey() */
#endif

#include "netgen.h"
#include "timing.h"
#include "hash.h"
#include "objlist.h"
#include "query.h"
#include "netfile.h"
#include "print.h"
#include "dbug.h"



/*************************************************************************/
/*                                                                       */
/*    I/O support for Query() routine                                    */
/*                                                                       */
/*************************************************************************/

static int SuppressPrompts = 0;
static char InputLine[200];

void typeahead(char *str)
{
  if (strlen(str) + strlen(InputLine) + 3 < sizeof(InputLine)) {
    strcat(InputLine," ");
    strcat(InputLine,str);
  }
  else fprintf(stderr, "InputLine too long: ignored command '%s'\n",str);
}

/* change the following to redirect the input stream */
FILE *promptstring_infile = NULL;

void promptstring(char *prompt, char *buf)
/* tries to get a token out of 'line' variable, 
but reads from 'promptstring_infile' if nec. */
/* copy it to buffer, when found */
/* If interactive, puts out 'prompt' */
{
  char *nexttok;
  char tmpstr[200];
  int echo;

  if (promptstring_infile == NULL)
    promptstring_infile = stdin;

  if (!SuppressPrompts) {
    Printf("%s",prompt);
    Fflush(stdout);
  }
  echo = 1;  /* assume we got it from typeahead */
  nexttok = InputLine;
  while (isspace(*nexttok) && *nexttok != '\0') nexttok++;
  if (*nexttok == '\0') {
    fgets(InputLine, sizeof(InputLine), promptstring_infile);
    if (promptstring_infile == stdin) echo = 0;
    nexttok = InputLine;
    while (isspace(*nexttok) && *nexttok != '\0') nexttok++;
    if (*nexttok == '\0') {
      *buf = '\0';
      return;
    }
  }
  /* nexttok points to beginning of valid token */
  strcpy(tmpstr,nexttok);
  nexttok = tmpstr;
  while (*nexttok != '\0' && !isspace(*nexttok)) nexttok++;
  strcpy(InputLine, nexttok);
  *nexttok = '\0';
  strcpy(buf, tmpstr);
  if (echo && !SuppressPrompts) Printf("%s\n",buf);
}


/* void InitializeCommandLine(int argc, char *argv[]) */
void InitializeCommandLine(int argc, char **argv)
{
  /* neither of the two Inits below are strictly necessary,
     as static objects are initialized to 0 (NULL) automatically    */
  InitCellHashTable();  
  InitGarbageCollection();
	
  /* generic transistor cells */
  CellDef("n");
    Primitive();
    Port("gate");
    Port("drain");
    Port("source");
  CellDef("p");
    Primitive();
    Port("gate");
    Port("drain");
    Port("source");
  CellDef("e");
    Primitive();
    Port("top");
    Port("bottom_a");
    Port("bottom_b");
  CellDef("b");
    Primitive();
    Port("base");
    Port("collector");
    Port("emitter");
  CellDef("r");
    Primitive();
    Port("r_dummy");
    Port("r_top");
    Port("r_bot");
  CellDef("c");
    Primitive();
    Port("c_top");
    Port("c_bot");
    Port("c_dummy");

#ifdef TCL_NETGEN
  return;

#else
  if (argv == NULL) return;	/* Don't run Query() */
		
  /* initialize command-line parser, including dbug code */
  DBUG_PROCESS(argv[0]);
  if (argc > 1) {
    int start;
    int usekbd, forceinteractive;
    usekbd = 1;   /* assume we are interactive */
    forceinteractive = 0; /* try to figure it out from cmd line */

    /* quiet prompting if program name is not netgen */
    SuppressPrompts =  (strstr(argv[0], "netgen") == NULL);

    for (start = 1; start < argc; start++) {
      if (argv[start][0] == '-') {
	switch (argv[start][1]) {
	case '#': DBUG_PUSH(&(argv[start][2])); 
	  break;
	case '\0': forceinteractive = 1; 
	  break;
	default: Fprintf(stderr,"Unrecognized switch: %s\n",argv[start]);
	  break;
	}
      }
      else {
	      typeahead(argv[start]);
	      usekbd = 0;
      }
    }
    if (!usekbd && !forceinteractive) typeahead("Q"); /* exit when done */
  }

  /* permit command-line typeahead even for X-windows */
  if (getenv("DISPLAY")) {
    int oldSuppressPrompts = SuppressPrompts;
    SuppressPrompts = 1;
    typeahead("q");  /* get out of one level of query */
    Query();
    SuppressPrompts = oldSuppressPrompts;
  }

#endif  /* TCL_NETGEN */
}

void Initialize(void)
{
#ifdef HAVE_MALLINFO
  char *wasted;

  wasted = (char *)MALLOC(2);  /* need to initialize memory allocator */
#endif

  InitializeCommandLine(0, NULL);
}


/* Print the type of object (mostly diagnostic) */

void PrintObjectType(int type)
{
   switch(type) {
      case UNIQUEGLOBAL:
	Printf("Unique Global");
	break;
      case GLOBAL:
	Printf("Global");
	break;
      case PORT:
	Printf("Port");
	break;
      case NODE:
	Printf("Node");
	break;
      default:
	if (type < 0) 
	   Printf("Error!");
	else
	   Printf("Pin %d", type);
	break;
   }
}
	


/*************************************************************************/
/*                                                                       */
/*    Some convenient routines for printing internal data structures     */
/*                                                                       */
/*************************************************************************/

#ifndef TCL_NETGEN

/*--------------------------------------------------------------*/
/* generate and print a list of elements that match a regexp	*/
/*--------------------------------------------------------------*/

void PrintElement(char *cell, char *list_template)
{
	
  struct objlist *list;
	
  if (strlen(cell))  CurrentCell = LookupCell(cell);
  list = List(list_template);
  Printf("Elements matching template: %s\n",list_template);
  while (list != NULL) {
    Printf ("   %s\n",list->name);
    list = list->next;
  }
}

#else

/*--------------------------------------------------------------*/
/* PrintElement() for use with Tcl---return all elements,	*/
/* let Tcl sort them out.					*/
/*--------------------------------------------------------------*/

void PrintAllElements(char *cell)
{
  struct nlist *np;
  struct objlist *ob;
  char *sfx;

  if (*cell == '\0') np = CurrentCell;
  else np = LookupCell(cell);
	
  if (np == NULL) {
    Printf("Cell '%s' not found.\n",cell);
    return;
  }
	
  ob = np->cell;
  for (ob = np->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
       if ((sfx = strrchr(ob->name, '/')) != NULL) *sfx = '\0';
       Printf("%s\n", ob->name);
       if (sfx != NULL) *sfx = '/';
    }
  }
}

#endif

/*--------------------------------------------------------------*/
/* Print connectivity between objects belonging to a specific	*/
/* node.  'filter' may be used to restrict the returned list to	*/
/* a specific type of object (node, element, port, pin, etc.).	*/
/*--------------------------------------------------------------*/

void Fanout(char *cell, char *node, int filter)
{
  struct nlist *np;
  struct objlist *ob;
  int nodenum;

  if (*cell == '\0') np = CurrentCell;
  else np = LookupCell(cell);
	
  if (np == NULL) {
    Printf("Cell '%s' not found.\n",cell);
    return;
  }
	
  nodenum = -999;
  for (ob = np->cell; ob != NULL; ob = ob->next) {
    if (match(node, ob->name)) {
      nodenum = ob->node;
      break;
    }
  }

  /* now print out all elements that connect to that node */

  if (nodenum == -999)
    Printf("Node '%s' not found in cell '%s'.\n", node, cell);
  else if (nodenum < 0)
    Printf("Node '%s' is disconnected.\n", node);
  else {
    if (ob != NULL)
       PrintObjectType(ob->type);
    else
       Printf("Object");
    Printf (" '%s' in cell '%s' connects to:\n", node, cell);
    ob = np->cell;
    while (ob != NULL) {
      if (ob->node == nodenum)
	 if (filter == ALLOBJECTS) {
	   Printf("  %s (", ob->name);
	   PrintObjectType(ob->type);
	   Printf(")\n");
	 }
	 else if ((filter == ALLELEMENTS) && (ob->type >= FIRSTPIN)) {
	   Printf("  %s\n", ob->name);
	 }
	 else if (ob->type == filter) {
	   Printf("  %s\n", ob->name);
	 }
      ob = ob->next;
    }
  }
}
	
#ifdef TCL_NETGEN

/* Print the nodes connected to each pin of the specified element */

void ElementNodes(char *cell, char *element)
{
  struct nlist *np;
  struct objlist *ob, *nob;
  int ckto;

  if (*cell == '\0') np = CurrentCell;
  else np = LookupCell(cell);
	
  if (np == NULL) {
    Printf("Cell '%s' not found.\n",cell);
    return;
  }

  ckto = strlen(element);
  for (ob = np->cell; ob != NULL; ob = ob->next) {
    if (!strncmp(element, ob->name, ckto))
       break;
  }
  if (ob == NULL) {
    Printf("Element '%s' not found in cell '%s'.\n", element, cell);
    return;
  }

  Printf("Element '%s' Pins:\n", element);
  for (; ob != NULL; ob = ob->next) {
    if (!strncmp(element, ob->name, ckto)) {
       Printf("   ");
       PrintObjectType(ob->type);
       Printf(" (%s)", ob->name + ckto + 1);
       for (nob = np->cell; nob != NULL; nob = nob->next) {
	 if (nob->type == NODE && nob->node == ob->node) {
	   Printf(" = %s\n", nob->name);
	 }
       }
    }
  }
  Printf("\n");
}

#endif  /* TCL_NETGEN */

/*--------------------------------------------------------------------*/
/* print all nodes in cell 'name', together with their connectivities */
/*--------------------------------------------------------------------*/

typedef struct _noderecord {
  char *name;					/* node name */
  int uniqueglobal, global, port, node, pin;	/* counts */
} noderecord;
  
void PrintNodes(char *name)
{
  int nodenum, nodemax;
  struct nlist *tp;
  struct objlist *ob;
  int maxnamelen;
  noderecord *nodelist;
  int uniqueglobals, globals, ports, nodes, pins;
	
  tp = LookupCell(name);
  if (tp == NULL) {
    Printf ("No cell '%s' found.\n",name);
    return;
  }
  Printf("Cell: '%s'\n",tp->name);
  
  nodemax = 0;
  maxnamelen = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    int len = strlen(NodeAlias(tp, ob));
    if (len > maxnamelen) maxnamelen = len;
    if (ob->node > nodemax) nodemax = ob->node;
  }
  nodelist = (noderecord *) CALLOC((nodemax + 1), sizeof(noderecord));
  
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    nodenum = ob->node;
    /* repeat bits of objlist.c here for speed */
#ifdef CACHE_NODENAMES
    if (tp->nodename_cache != NULL) {
      nodelist[nodenum].name = tp->nodename_cache[nodenum]->name;
    }
    else {
#endif
      /* Overwrite name, in order of precedence */
      if ((nodelist[nodenum].port == 0) &&
	  ((ob->type == PORT) ||
	  ((nodelist[nodenum].node == 0) &&
	  ((ob->type == NODE) ||
	  ((nodelist[nodenum].uniqueglobal == 0) &&
	  ((ob->type == UNIQUEGLOBAL) ||
	  ((nodelist[nodenum].global == 0) &&
	  ((ob->type == GLOBAL) ||
	  ((nodelist[nodenum].pin == 0) &&
	  (ob->type >= FIRSTPIN))))))))))
	nodelist[nodenum].name = ob->name;

#ifdef CACHE_NODENAMES
    }
#endif
    switch (ob->type) {
      case UNIQUEGLOBAL:
	nodelist[nodenum].uniqueglobal++; break;
      case GLOBAL:
	nodelist[nodenum].global++; break;
      case PORT:
	nodelist[nodenum].port++; break;
      case NODE:
	nodelist[nodenum].node++; break;
      default:
	nodelist[nodenum].pin++; break;
    }
  }

  for (nodenum = 0; nodenum < nodemax; nodenum++) {
    if (nodelist[nodenum].name == NULL) continue;

    pins = nodelist[nodenum].pin;
    ports = nodelist[nodenum].port;
    globals = nodelist[nodenum].global;
    uniqueglobals = nodelist[nodenum].uniqueglobal;
    nodes = nodelist[nodenum].node;

    Printf("Node %d (%s):", nodenum, nodelist[nodenum].name);
    Ftab(stdout, maxnamelen + 15);
    Printf("Total = %d,", pins + ports + nodes + globals + uniqueglobals);
    if (ports)
      Printf(" Ports = %d,", ports);
    Ftab(stdout, maxnamelen + 40);
    if (pins)
      Printf("Pins = %d,", pins);
    Ftab(stdout, maxnamelen + 52);
    if (nodes)
      Printf("Nodes = %d,", nodes);
    Ftab(stdout, maxnamelen + 63);
    if (globals)
      Printf("Globals = %d,", globals);
    Ftab(stdout, maxnamelen + 80);
    if (uniqueglobals)
      Printf("UniqueGlobals = %d", uniqueglobals);
    Printf("\n");
  }

  FREE(nodelist);
}
  
void PrintCell(char *name)
{
  struct nlist *tp;
  struct objlist *ob;
  int maxnamelen;
	
  tp = LookupCell(name);
  if (tp == NULL) {
    Printf ("No cell '%s' found.\n",name);
    return;
  }
  maxnamelen = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
	  int len;
	  if ((len = strlen(ob->name)) > maxnamelen) maxnamelen = len;
  }
  
  Printf("Cell: '%s'\n",tp->name);
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    Printf ("'%s' ",ob->name);
    Ftab(stdout, maxnamelen + 4);
    switch (ob->type) {
      case UNIQUEGLOBAL:	
	Printf("unique global"); break;
      case GLOBAL:
	Printf("global"); break;
      case PORT:
	Printf("port"); break;
      case NODE:
	Printf("node"); break;
      default:
	Printf("(class = %s)",ob->classname);
        Ftab(stdout, 40);
	Printf("pin %d", ob->type);
	break;
    }
    Ftab(stdout, 50);
    Printf(" Electrical node: %d\n",ob->node);
  }
}

void PrintInstances(char *name)
{
  struct nlist *tp;
  struct objlist *ob;
  int instancecount;
	
  tp = LookupCell(name);
  if (tp == NULL) {
    Printf ("No cell '%s' found.\n",name);
    return;
  }
  Printf("Cell: '%s'\n",tp->name);
  instancecount = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
      struct objlist *ob2;
      int port, node, global, uniqueglobal, pin;
      int ports, nodes, globals, uniqueglobals, pins;

      port = node = global = uniqueglobal = pin = 0;
      instancecount++;

      ob2 = ob;
      do {
	struct objlist *ob3;

	ports = nodes = globals = uniqueglobals = pins = 0;
	for (ob3 = tp->cell; ob3 != NULL; ob3 = ob3->next)
	  if (ob3->node == ob2->node)
	    switch (ob3->type) {
	    case UNIQUEGLOBAL: uniqueglobals++; break;
	    case GLOBAL: globals++; break;
	    case PORT:   ports++; break;
	    case NODE:   nodes++; break;
	    default:     pins++; break;
	    }
	pin++;
	if (uniqueglobals) uniqueglobal++;
	else if (globals) global++;
	else if (ports) port++;
	else if (nodes) node++;
	ob2 = ob2->next;
      } while (ob2 != NULL && ob2->type > FIRSTPIN);
/*      Fflush(stdout); */
      Printf("%s (class: %s)", ob->instance, ob->classname);
      Ftab(stdout,35);
      Printf("%2d pins ->",pin);
      if (port) Printf("%2d ports,",port);
      Ftab(stdout,55);
      if (node) Printf("%2d nodes,",node);
      Ftab(stdout,65);
      if (global) Printf("%2d globals,",global);
      Ftab(stdout,75);
      if (uniqueglobal) Printf("%2d ug",uniqueglobal);
      Printf("\n");
    }
  }
  Printf("Cell %s contains %d instances.\n",name,instancecount);
}

void DescribeInstance(char *name)
{
  struct nlist *tp;
  struct objlist *ob;
  unsigned char *instlist;
  int nodemax, nodenum;

  int instancecount;
  int node, nodenumber, morenodes, disconnectednodes;
	
  tp = LookupCell(name);
  if (tp == NULL) {
    Printf ("No cell '%s' found.\n",name);
    return;
  }
  Printf("Cell: '%s'\n",tp->name);

  /* First pass counts total number of entries */
  nodemax = 0;
  disconnectednodes = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next) 
  {
    if (ob->node > nodemax) nodemax = ob->node;
    else if (ob->node == -1) {
      disconnectednodes++;
      Printf("  disconnected node: %s\n", ob->name);
    }
  }
  instlist = (unsigned char *) CALLOC((nodemax + 1), sizeof(unsigned char));

  /* Second pass finds total number of unique nodes */
  for (ob = tp->cell; ob != NULL; ob = ob->next) 
    if (ob->node > 0)
      instlist[ob->node] = (unsigned char) 1;

  /* Now add them all together */
  nodenumber = 0;
  for (nodenum = 1; nodenum <= nodemax; nodenum++)
     if (instlist[nodenum] == (unsigned char) 1) nodenumber++;

  /* And we're done with this record */
  FREE(instlist);

  /* Now collect the relevant information per node */

  ClearDumpedList();

  instancecount = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
      instancecount++;
      (LookupCell(ob->classname)->dumped)++;
    }
  }
  Printf("Cell %s contains %d instances.\n",name,instancecount);

  /* print out results */
  tp = FirstCell();
  while (tp != NULL) {
    if (tp->dumped) {
      Printf("  Class: %s", tp->name);
      Ftab(stdout, 30);
      Printf(" instances: %3d\n", tp->dumped);
    }
    tp = NextCell();
  }
  Printf("Cell contains %d nodes", nodenumber);
  if (disconnectednodes) 
    Printf(", and %d disconnected pins", disconnectednodes);
  Printf(".\n");

}

void PrintPortsInCell(char *cellname)
{
  struct nlist *np;
  struct objlist *ob;
  int portcount;

  np = LookupCell(cellname);
  if (np == NULL) {
    Printf("No cell: %s\n",cellname);
    return;
  }
  portcount = 0;
  for (ob = np->cell; ob != NULL; ob = ob->next) 
    if (IsPort(ob->type)) {
      portcount++;
      Printf("%s\n", ob->name);
    }
  Printf("Cell %s contains %d ports.\n",cellname, portcount);
}

void PrintLeavesInCell(char *cellname)
{
  struct nlist *np;
  struct objlist *ob;
  int am_a_leaf;

  np = LookupCell(cellname);
  if (np == NULL) {
    Printf("No cell: %s\n",cellname);
    return;
  }
  if (np->dumped) return;
  np->dumped = 1;

  if (np->primitive) {
    Printf("%s; %d ports; Primitive.\n", cellname, NumberOfPorts(cellname));
    return;
  }

  /* otherwise, consider it's children */
  am_a_leaf = 1; /* assume I am a leaf */
  for (ob = np->cell; ob != NULL; ob = ob->next)
    if (ob->type == FIRSTPIN) {
      /* if I contain an instance, I cannot be a leaf */
      PrintLeavesInCell(ob->classname);
      am_a_leaf = 0;
    }

  if (am_a_leaf) Printf("%s; %d ports\n", cellname, NumberOfPorts(cellname));
  return;
}

static int PrintLeavesInCellHash(struct hashlist *p)
/* print leaves in hash table that are INSTANCED by other cells */
{
  struct nlist *ptr;

  ptr = (struct nlist *)(p->ptr);
  if (!(ptr->primitive)) PrintLeavesInCell(ptr->name);
  return(0);
}

void PrintAllLeaves(void)
{
  ClearDumpedList();
  RecurseCellHashTable(PrintLeavesInCellHash);
}

void TogglePrimitive(char *cellname)
{
  struct nlist *np;

  np = LookupCell(cellname);
  if (np == NULL) {
    Printf("No cell: %s\n",cellname);
    return;
  }
  np->primitive = !(np->primitive);
  Printf("Cell '%s' is%s primitive.\n", 
	 cellname, (np->primitive)?" now":" no longer");
}

void ToggleAllPrimitive(void)
{
  struct nlist *p;
  int dat;

  p = FirstCell();
  if (p != NULL) {
    dat = !(p->primitive);
    while (p != NULL) {
      p->primitive = dat;
      p = NextCell();
    }
    if (dat) Printf("All cells are now Primitive\n");
    else Printf("No Primitive cells remain.\n");
  }
}

static jmp_buf jmpenv;

/* static void handler(void) */
static void handler(int sig)
{
  fprintf(stderr,"\nInterrupt (%d)!!\n", sig);
  fflush(stderr);
  longjmp(jmpenv,1);
}

#ifndef TCL_NETGEN

void Query(void)
{
  /* little interactive debugger */
  char reply;
  char repstr[100];
  char repstr2[100];
  float StartTime;   /* for elapsed CPU times */
  int Timing;  /* if true, print times of each command */
	
  if (!SuppressPrompts)
    Printf("Netgen %s.%s: %s, %s, %s\n", NETGEN_VERSION, NETGEN_REVISION,
		NETGEN_COPYRIGHT, NETGEN_AUTHOR, NETGEN_DEVELOPER);
  setjmp(jmpenv);
#if defined(CPLUSPLUS) && defined(SIG_PF)
  signal(SIGINT,(SIG_PF)handler);  /* patch by Glenn */
#else
  signal(SIGINT,handler);
#endif
  Timing = 0;
  StartTime = 0.0;
  do {
    promptstring("NETGEN command: ",repstr);
    if (Timing) StartTime = CPUTime();
    reply = repstr[0];
    switch (reply) {
    case 'h' : PrintCellHashTable(0); break;
    case 'H' : PrintCellHashTable(1); break;
    case 'N' :
      promptstring("Enter cell name: ", repstr);
      PrintNodes(repstr);
      break;
    case 'n' : 
      promptstring("Enter element name: ", repstr);
      if (CurrentCell == NULL) 
	promptstring("Enter cell name:    ", repstr2);
      else strcpy(repstr2, CurrentCell->name);
      Fanout(repstr2, repstr, ALLOBJECTS);
      break;
    case 'e' : 
      promptstring("Enter element name: ", repstr);
      if (CurrentCell == NULL) 
	promptstring("Enter cell name:    ", repstr2);
      else strcpy(repstr2, CurrentCell->name);
      PrintElement(repstr2,repstr);
      break;
    case 'c' : 
      promptstring("Enter cell name: ", repstr);
      PrintCell(repstr);
      break;
    case 'i' :
      promptstring("Enter cell name: ", repstr);
      PrintInstances(repstr);
      break;
    case 'd':
      promptstring("Enter cell name: ", repstr);
      DescribeInstance(repstr);
      break;
    case 'v' : 
      promptstring("Enter cell name: ", repstr);
      TogglePrimitive(repstr);
      break;
    case 'V' : 
      ToggleAllPrimitive();
      break;
/* output file formats */
    case 'k' : 
      promptstring("Write NTK: Enter cell name: ", repstr);
      Ntk(repstr,"");
      break;
    case 'x' : 
      promptstring("Write EXT: Enter cell name: ", repstr);
      Ext(repstr);
      break;
    case 'z' : 
      promptstring("Write SIM: Enter cell name: ", repstr);
      Sim(repstr);
      break;
    case 'w' : 
      promptstring("Write WOMBAT: cell name: ", repstr);
      Wombat(repstr,NULL);
      break;
    case 'a' : 
      promptstring("Write ACTEL: cell name: ", repstr);
      Actel(repstr,"");
      break;
    case 's':
      promptstring("Write SPICE: cell name: ", repstr);
      SpiceCell(repstr,"");
      break;
    case 'E':
      promptstring("Write ESACAP: cell name: ", repstr);
      EsacapCell(repstr,"");
      break;
    case 'g':
      promptstring("Write NETGEN: cell name: ", repstr);
      WriteNetgenFile(repstr,"");
      break;
    case 'C':
      promptstring("Write C code: cell name: ", repstr);
      Ccode(repstr,"");
      break;
/* input file formats */
    case 'r':
    case 'R':
      promptstring("Read file: ",repstr);
      ReadNetlist(repstr);
      break;
    case 'X' : 
      promptstring("Read EXT: Enter file name: ", repstr);
      ReadExtHier(repstr);
      break;
    case 'Z' : 
      promptstring("Read SIM: Enter file name: ", repstr);
      ReadSim(repstr);
      break;
    case 'K' : 
      promptstring("Read NTK: file? ", repstr);
      ReadNtk(repstr);
      break;
    case 'A' : 
      printf("Reading ACTEL library.\n");
      ActelLib();
      break;
    case 'S':
      promptstring("Read SPICE (.ckt) file? ", repstr);
      ReadSpice(repstr);
      break;
    case 'G' : 
      promptstring("Read NETGEN: file? ", repstr);
      ReadNetgenFile(repstr);
      break;

    case 'f' : 
      promptstring("Enter cell name to flatten: ", repstr);
      Flatten(repstr);
      break;
    case 'F' : 
      promptstring("Enter class of cells to flatten: ", repstr);
      FlattenInstancesOf(repstr);
      break;
    case 'p' : 
      promptstring("Enter cell name: ", repstr);
      PrintPortsInCell(repstr);
      break;
    case 'T' : 
      NETCOMP();
      break;
    case 'l' : 
      ClearDumpedList();
      promptstring("Enter cell name: ", repstr);
      PrintLeavesInCell(repstr);
      break;
    case 'L' : 
      printf("List of all leaf cells:\n");
      PrintAllLeaves();
      break;
    case 'D' : Debug = !Debug; 
      printf("Debug mode is %s\n", Debug?"ON":"OFF");
      break;
    case 't': StartTime = CPUTime();
      Timing = !Timing;
      printf("Timing of commands %s.\n", Timing?"enabled":"disabled");
      break;
    case '#':
      if (strlen(repstr) > 1) {
	char *command;
	command = repstr+1;
	DBUG_PUSH(command);
      }
      else {
	promptstring("Dbug command? ",repstr);
	DBUG_PUSH(repstr);
      }
      break;
    case '!': 
#ifdef IBMPC
      system("");
#else
      system("/bin/csh");
#endif
      break;
    case 'I' : Initialize(); break;
    case 'q' : break;
    case 'Q' : exit(0);
    case 'P' : PROTOCHIP(); break;
#ifdef HAVE_MALLINFO
    case 'm': PrintMemoryStats(); break;
#endif
    case '<' :
      {
	FILE *oldfile;
	promptstring("Read from file? ",repstr);
	oldfile = promptstring_infile;
	promptstring_infile = fopen(repstr,"r");
	if (promptstring_infile == NULL) 
	  Printf("Unable to open command file: %s\n", repstr);
	else {
	  Query();
	  fclose(promptstring_infile);
	}
	promptstring_infile = oldfile;
	break;
      }
    default : 
      fflush(stdout);
      if (strlen(repstr)) fprintf(stderr,"Unknown command: %c\n",reply);
      fflush(stderr);
      printf(
"WRITE: nt(k),e(x)t,sim (z),(w)ombat,(a)ctel,net(g)en,(s)pice,(E)sacap,(C)\n");
      printf(
"READ:  (r)ead, nt(K), e(X)t, sim (Z), (A)ctel library, net(G)en, (S)pice\n");
      printf(
"Print LISTS: (d)escribe cell (c)ontents, (i)nstances, (p)orts, (l)eaves (L)\n"
	    );
      printf(
"PRINT: (n)ode connectivity (N), (e)lement list, (h)ash table entries (H))\n");
      printf(
"CELL OPS: (f)latten cell, (F)latten instance, make cell primiti(v)e (V=all)\n"
	    );

      printf("toggle (D)ebug, (t)ime commands, embed (P)rotochip, ne(T)cmp\n");
      printf("(!) push shell, (<) read input file");

#ifdef HAVE_MALLINFO
      printf(", show (m)emory usage");
#endif
      printf("\n");
      printf("(q)uit, (Q)uit immediately, re-(I)nitialize \n");
      break;
    }
    if (Timing) {
      printf("CPU time used by last operation = %0.2f s\n", 
	     ElapsedCPUTime(StartTime));
      fflush(stdout); /* just in case we have been redirected */
    }
  } while (reply != 'q');
#if defined(CPLUSPLUS) && defined(SIG_PF)
  signal(SIGINT,(SIG_PF)SIG_DFL);
#else
  signal(SIGINT,SIG_DFL);
#endif
}

#endif
