/* "NETGEN", a netlist-specification tool for VLSI
   Copyright (C) 1989, 1990   Massimo A. Sivilotti
   Author's address: mass@csvax.cs.caltech.edu;
                     Caltech 256-80, Pasadena CA 91125.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (any version).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file copying.  If not, write to
the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* spice.c -- Input / output for SPICE and ESACAP formats */

#include "config.h"

#include <stdio.h>
#if 0
#include <stdarg.h>  /* what about varargs, like in pdutils.c ??? */
#endif

#ifdef IBMPC
#include <stdlib.h>  /* for calloc(), free() */
#endif

#include "netgen.h"
#include "hash.h"
#include "objlist.h"
#include "netfile.h"
#include "print.h"


void SpiceSubCell(struct nlist *tp, int IsSubCell)
{
  struct objlist *ob;
  int node, maxnode;

  /* check to see that all children have been dumped */
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
      struct nlist *tp2;

      tp2 = LookupCell(ob->classname);
      if ((tp2 != NULL) && !(tp2->dumped) && !(tp2->primitive)) 
	SpiceSubCell(tp2, 1);
    }
  }

  /* print preface, if it is a subcell */
  if (IsSubCell) {
    FlushString(".SUBCKT %s ",tp->name);
    for (ob = tp->cell; ob != NULL; ob = ob->next) 
      if (IsPortInPortlist(ob, tp)) FlushString("%d ", ob->node);
    FlushString("\n");
  }

  /* print names of all nodes, prefixed by comment character */
  maxnode = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next)
    if (ob->node > maxnode) maxnode = ob->node;

/* was:  for (node = 0; node <= maxnode; node++)  */
  for (node = 1; node <= maxnode; node++) 
    FlushString("# %3d = %s\n", node, NodeName(tp, node));

  /* traverse list of objects */
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
      int drain_node, gate_node, source_node;
      /* print out element, but special-case transistors */
      if (match (ob->classname, "n") || match(ob->classname, "p")) {
	if (ob->instance[0] == 'M')
	  FlushString("%s ",ob->instance);
	else
	  FlushString("M%s ",ob->instance);
	/* note: this code is dependent on the order defined in Initialize()*/
	gate_node = ob->node;
	ob = ob->next;
	drain_node = ob->node;
	ob = ob->next;
	source_node = ob->node;
	FlushString("%d %d %d ",drain_node, gate_node, source_node);
	/* write fake substrate connections: NSUB and PSUB */
	/* write fake transistor classes: NTRAN and PTRAN  */
	if (match (ob->classname, "n")) FlushString("NSUB NTRAN\n");
	else FlushString("PSUB PTRAN\n");
      }
      else {
	/* it must be a subckt */
	FlushString("X%s %d ", ob->instance, ob->node);
	while (ob->next != NULL && ob->next->type > FIRSTPIN) {
	  ob = ob->next;
	  FlushString("%d ",ob->node);
	}
	FlushString("X%s\n", ob->classname);
      }
    }
  }
	
  if (IsSubCell) FlushString(".ENDS\n");
  tp->dumped = 1;
}


void SpiceCell(char *name, char *filename)
{
  struct nlist *tp;
  char FileName[500];

  tp = LookupCell(name);
  if (tp == NULL) {
    Printf ("No cell '%s' found.\n", name);
    return;
  }

  if (filename == NULL || strlen(filename) == 0) 
    SetExtension(FileName, name, SPICE_EXTENSION);
  else 
    SetExtension(FileName, filename, SPICE_EXTENSION);

  if (!OpenFile(FileName, 80)) {
    perror("ext(): Unable to open output file.");
    return;
  }
  ClearDumpedList();
  /* all spice decks begin with comment line */
  FlushString("SPICE deck for cell %s written by Netgen %s.%s\n\n", 
	      name, NETGEN_VERSION, NETGEN_REVISION);
  SpiceSubCell(tp, 0);
  CloseFile(FileName);
}

char *ReadSpice(char *fname)
{
  int cdnum = 1, rdnum = 1, ndev;
  int warnings = 0;
  int CellDefInProgress = 0;
  
  if (!OpenParseFile(fname)) {
    char name[100];

    SetExtension(name, fname, SPICE_EXTENSION);
    if (!OpenParseFile(name)) {
      Fprintf(stderr,"No file: %s\n",name);
      return NULL;
    }    
  }

  /* all spice files start with a comment line */
  SkipTok();
  SkipNewLine();

  while (!EndParseFile()) {
    SkipTok(); /* get the next token */
    if (EndParseFile()) break;

    if (nexttok[0] == '*') SkipNewLine();

    else if (toupper(nexttok[0]) == 'Q') {
      char inst[100], emitter[100], base[100], collector[100], tmp[100];

      if (!CellDefInProgress) {
	/* fake declaration of top-level cell */
	CellDef(fname);
	CellDefInProgress = 1;
      }
      strcpy(inst, nexttok); SkipTok(); 
      strcpy(collector, nexttok); SkipTok();
      strcpy(base, nexttok);   SkipTok();
      strcpy(emitter, nexttok);  SkipTok();
      /* make sure all the nodes exist */
      if (LookupObject(collector, CurrentCell) == NULL) Node(collector);
      if (LookupObject(base, CurrentCell) == NULL) Node(base);
      if (LookupObject(emitter, CurrentCell) == NULL) Node(emitter);

      ndev = 1;
      while (nexttok != NULL)
      {
	 /* Parse for W, L, and M parameters */

	 SkipTokNoNewline();
	 if ((nexttok == NULL) || (nexttok[0] == '\0')) break;
	 if (nexttok[1] == '=')
	 {
	    switch (toupper(nexttok[0]))
	    {
		case 'M':
		    sscanf(&nexttok[2], "%d", &ndev);
		    break;
		case 'W':
		    /* No action defined, yet */
		    break;
		case 'L':
		    /* No action defined, yet */
		    break;
	    }
	 }
      }
      while (ndev > 0)
      {
	 B(NULL, base, collector, emitter);
	 ndev--;
      }
    }
    else if (toupper(nexttok[0]) == 'M') {
      char inst[100], drain[100], gate[100], source[100], tmp[100];

      if (!CellDefInProgress) {
	/* fake declaration of top-level cell */
	CellDef(fname);
	CellDefInProgress = 1;
      }
      strcpy(inst, nexttok); SkipTok(); 
      strcpy(drain, nexttok);  SkipTok();
      strcpy(gate, nexttok);   SkipTok();
      strcpy(source, nexttok); SkipTok();
      /* make sure all the nodes exist */
      if (LookupObject(drain, CurrentCell) == NULL) Node(drain);
      if (LookupObject(gate, CurrentCell) == NULL) Node(gate);
      if (LookupObject(source, CurrentCell) == NULL) Node(source);

      SkipTok(); /* skip the substrate, to get to the device model */
      switch(nexttok[0]) {
	 case 'N': case 'n':
	    ndev = 1;
	    while (nexttok != NULL)
	    {
		/* Parse for W, L, and M parameters */

		SkipTokNoNewline();
		if ((nexttok == NULL) || (nexttok[0] == '\0')) break;
		if (nexttok[1] == '=')
		{
		    switch (toupper(nexttok[0]))
		    {
			case 'M':
			    sscanf(&nexttok[2], "%d", &ndev);
			    break;
			case 'W':
			    /* No action defined, yet */
			    break;
			case 'L':
			    /* No action defined, yet */
			    break;
		    }
		}
	    }
	    while (ndev > 0)
	    {
		N(NULL, gate, drain, source);
		ndev--;
	    }
	    break;
	 case 'P': case 'p':
	    ndev = 1;
	    while (nexttok != NULL)
	    {
		/* Parse for W, L, and M parameters */

		SkipTokNoNewline();
		if ((nexttok == NULL) || (nexttok[0] == '\0')) break;
		if (nexttok[1] == '=')
		{
		    switch (toupper(nexttok[0]))
		    {
			case 'M':
			    sscanf(&nexttok[2], "%d", &ndev);
			    break;
			case 'W':
			    /* No action defined, yet */
			    break;
			case 'L':
			    /* No action defined, yet */
			    break;
		    }
		}
	    }
	    while (ndev > 0)
	    {
		P(NULL, gate, drain, source);
		ndev--;
	    }
	    break;
	 case 'Z': case 'z':	/* 3-port resistors from magic extractor */
	    Res(NULL, gate, source, drain);
	    break;
	 default:
	    Fprintf(stderr,"Unknown model type %s in instance %s; ignoring\n",
		nexttok, inst);
	    goto instskip;
	    break;
      }

instskip:
      SkipNewLine();
    }
    else if (toupper(nexttok[0]) == 'C') {	/* 2-port capacitors */
      if (IgnoreRC) {
	 SkipNewLine();
      }
      else {
        char inst[100], ctop[200], cbot[200], cdummy[200];

        if (!CellDefInProgress) {
	  /* fake declaration of top-level cell */
	  CellDef(fname);
	  CellDefInProgress = 1;
        }
        strcpy(inst, nexttok); SkipTok(); 
        strcpy(ctop, nexttok);  SkipTok();
        strcpy(cbot, nexttok);   SkipTok();
        /* make sure all the nodes exist */
        if (LookupObject(ctop, CurrentCell) == NULL) Node(ctop);
        if (LookupObject(cbot, CurrentCell) == NULL) Node(cbot);
        SkipNewLine(); /* skip capacitor value and possible attributes */

        sprintf(cdummy, "cdum%d", cdnum++);
        Node(cdummy);   /* define the (unique) node */
        Cap(NULL, ctop, cbot, cdummy);
      }
    }
    else if (toupper(nexttok[0]) == 'R') {	/* 2-port resistors */
      if (IgnoreRC) {
	 SkipNewLine();
      }
      else {
        char inst[100], rtop[200], rbot[200], rdummy[200];

        if (!CellDefInProgress) {
	  /* fake declaration of top-level cell */
	  CellDef(fname);
	  CellDefInProgress = 1;
        }
        strcpy(inst, nexttok); SkipTok(); 
        strcpy(rtop, nexttok);  SkipTok();
        strcpy(rbot, nexttok);   SkipTok();
        /* make sure all the nodes exist */
        if (LookupObject(rtop, CurrentCell) == NULL) Node(rtop);
        if (LookupObject(rbot, CurrentCell) == NULL) Node(rbot);
        SkipNewLine(); /* skip capacitor value and possible attributes */

        sprintf(rdummy, "rdum%d", rdnum++);
        Node(rdummy);   /* define the (unique) node */
        Res(NULL, rdummy, rtop, rbot);
      }
    }
    else if (toupper(nexttok[0]) == 'X') {	/* subcircuit instances */
      char instancename[100], subcktname[100];

      struct portelement {
	char *name;
	struct portelement *next;
      };

      struct portelement *head, *tail, *scan, *scannext;
      struct objlist *obptr;

      strcpy(instancename, nexttok + 1);
      if (!CellDefInProgress) {
	/* fake declaration of top-level cell */
	CellDef(fname);
	CellDefInProgress = 1;
      }
      
      head = NULL;
      tail = NULL;
      SkipTok();
      while (nexttok != NULL) {
	/* must still be a node */
	struct portelement *new_port;

	new_port = (struct portelement *)CALLOC(1, sizeof(struct portelement));
	new_port->name = strsave(nexttok);
	if (head == NULL) head = new_port;
	else tail->next = new_port;
	new_port->next = NULL;
	tail = new_port;
	SkipTokNoNewline();
      }


      /* find the last element of the list, which is not a port,
         but the class type */
      scan = head;
      while (scan != NULL && scan->next != tail) scan = scan->next;
      tail = scan;
      scan = scan->next;
      tail->next = NULL;

      strcpy(subcktname, scan->name);
      free (scan->name);
      FREE (scan);

      /* nexttok is now NULL, scan->name points to class */
      Instance(subcktname, instancename);

      /* (Diagnostic) */
      /* Fprintf(stderr, "instancing subcell: %s (%s):", subcktname, instancename); */
      /*
      for (scan = head; scan != NULL; scan = scan->next)
	Fprintf(stderr," %s", scan->name);
      Fprintf(stderr,"\n");
      */
      
      obptr = LookupInstance(instancename, CurrentCell);
      if (obptr != NULL) {
        scan = head;
        do {
	  if (LookupObject(scan->name, CurrentCell) == NULL) Node(scan->name);
	  join(scan->name, obptr->name);
	  obptr = obptr->next;
	  scan = scan->next;
        } while (obptr != NULL && obptr->type > FIRSTPIN && scan != NULL);

        if ((obptr == NULL && scan != NULL) ||
		(obptr != NULL && scan == NULL && obptr->type > FIRSTPIN)) {
	  if (warnings <= 100) {
	     Fprintf(stderr,"Parameter list mismatch in %s :: %s\n",
		nexttok, instancename);
	     InputParseError(stderr);
	     if (warnings == 100)
	        Fprintf(stderr, "Too many warnings. . . will not report any more.\n");
          }
	  warnings++;
	}
      }

      /* free up the allocated list */
      scan = head;
      while (scan != NULL) {
	scannext = scan->next;
	FREE(scan->name);
	FREE(scan);
	scan = scannext;
      }
    }
    else if (matchnocase(nexttok, ".SUBCKT")) {
      SkipTok();
      CellDef(nexttok);
      CellDefInProgress = 1;
      /* tokens on the rest of the line are ports */
      SkipTokNoNewline();
      while (nexttok != NULL) {
	Port(nexttok);
	SkipTokNoNewline();
      }
    }
    else if (matchnocase(nexttok, ".ENDS")) {
      EndCell();
      CellDefInProgress = 0;
      SkipNewLine();
    }
    else if (matchnocase(nexttok, "+")) {
      if (!warnings) {
	 Fprintf(stderr,"Ignoring line starting with token: %s\n", nexttok);
	 InputParseError(stderr);
	 Fprintf(stderr,"This warning will not be repeated.\n");
      }
      warnings++;
      SkipNewLine();
    }
    else {
      if (warnings <= 100) {
	 Fprintf(stderr, "Ignoring line starting with token: %s\n", nexttok);
	 InputParseError(stderr);
	 if (warnings == 100)
	    Fprintf(stderr, "Too many warnings. . . will not report any more.\n");
      }
      warnings++;
      SkipNewLine();
    }
  }
  CloseParseFile();
  if (warnings)
     Fprintf(stderr, "File %s read with %d warning%s.\n", fname,
		warnings, (warnings == 1) ? "" : "s");
  return fname;
}


void EsacapSubCell(struct nlist *tp, int IsSubCell)
{
  struct objlist *ob;
  int node, maxnode;

  /* check to see that all children have been dumped */
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
      struct nlist *tp2;

      tp2 = LookupCell(ob->classname);
      if ((tp2 != NULL) && !(tp2->dumped) && !(tp2->primitive)) 
	EsacapSubCell(tp2, 1);
    }
  }

  /* print preface, if it is a subcell */
  if (IsSubCell) {
    FlushString("# %s doesn't know how to generate ESACAP subcells\n");
    FlushString("# Look in spice.c \n\n");
    FlushString(".SUBCKT %s ",tp->name);
    for (ob = tp->cell; ob != NULL; ob = ob->next) 
      if (IsPortInPortlist(ob, tp)) FlushString("%d ", ob->node);
    FlushString("# End of bogus ESACAP subcell\n");
    FlushString("\n");
  }

  /* print names of all nodes, prefixed by comment character */
  maxnode = 0;
  for (ob = tp->cell; ob != NULL; ob = ob->next)
    if (ob->node > maxnode) maxnode = ob->node;

/* was: for (node = 0; node <= maxnode; node++) */
  for (node = 1; node <= maxnode; node++) 
    FlushString("# %3d = %s\n", node, NodeName(tp, node));


  /* traverse list of objects */
  for (ob = tp->cell; ob != NULL; ob = ob->next) {
    if (ob->type == FIRSTPIN) {
      int drain_node, gate_node, source_node;
      /* print out element, but special-case transistors */
      if (match (ob->classname, "n") || match(ob->classname, "p")) {
	FlushString("X%s ",ob->instance);
	/* note: this code is dependent on the order defined in Initialize()*/
	gate_node = ob->node;
	ob = ob->next;
	drain_node = ob->node;
	ob = ob->next;
	source_node = ob->node;
	FlushString("(%d %d %d ",drain_node, gate_node, source_node);
	/* write fake substrate connections: NSUB and PSUB */
	/* write fake transistor sizes: NL, NW, PL and PW */
	/* write fake transistor classes: NCHANNEL and PCHANNEL  */
	if (match (ob->classname, "n")) 
	  FlushString("NSUB)=SMOS(TYPE=NCHANNEL,W=NW,L=NL);\n");
	else FlushString("PSUB)=SMOS(TYPE=PCHANNEL,W=PW,L=PL);\n");
      }
      else {
	/* it must be a subckt */
	FlushString("### BOGUS SUBCKT: X%s %d ", ob->instance, ob->node);
	while (ob->next != NULL && ob->next->type > FIRSTPIN) {
	  ob = ob->next;
	  FlushString("%d ",ob->node);
	}
	FlushString("X%s\n", ob->classname);
      }
    }
  }
	
  if (IsSubCell) FlushString(".ENDS\n");
  tp->dumped = 1;
}


void EsacapCell(char *name, char *filename)
{
  struct nlist *tp;
  char FileName[500];

  tp = LookupCell(name);
  if (tp == NULL) {
    Printf ("No cell '%s' found.\n", name);
    return;
  }

  if (filename == NULL || strlen(filename) == 0) 
    SetExtension(FileName, name, ESACAP_EXTENSION);
  else 
    SetExtension(FileName, filename, ESACAP_EXTENSION);

  if (!OpenFile(FileName, 80)) {
    perror("ext(): Unable to open output file.");
    return;
  }
  ClearDumpedList();
  /* all Esacap decks begin with the following comment line */
  FlushString("# ESACAP deck for cell %s written by Netgen %s.%s\n\n", 
	      name, NETGEN_VERSION, NETGEN_REVISION);
  EsacapSubCell(tp, 0);
  FlushString("# end of ESACAP deck written by Netgen %s.%s\n\n",
		NETGEN_VERSION, NETGEN_REVISION);
  CloseFile(FileName);
}
